using LinearAlgebra, Test, PreallocationTools, ForwardDiff, LabelledArrays,
      RecursiveArrayTools

function test(u0, dual, chunk_size)
    cache = PreallocationTools.dualcache(u0, chunk_size)
    allocs_normal1 = @allocated get_tmp(cache, u0)
    allocs_normal2 = @allocated get_tmp(cache, first(u0))
    allocs_dual1 = @allocated get_tmp(cache, dual)
    allocs_dual2 = @allocated get_tmp(cache, first(dual))
    result_normal1 = get_tmp(cache, u0)
    result_normal2 = get_tmp(cache, first(u0))
    result_dual1 = get_tmp(cache, dual)
    result_dual2 = get_tmp(cache, first(dual))
    return allocs_normal1, allocs_normal2, allocs_dual1, allocs_dual2, result_normal1,
           result_normal2, result_dual1,
           result_dual2
end

#Setup Base Array tests
chunk_size = 5
u0 = ones(5, 5)
dual = zeros(ForwardDiff.Dual{ForwardDiff.Tag{nothing, Float64}, Float64,
                              chunk_size}, 5, 5)
results = test(u0, dual, chunk_size)
#allocation tests
@test results[1] == 0
@test results[2] == 0
@test results[3] == 0
@test results[4] == 0
#size tests
@test size(results[5]) == size(u0)
@test size(results[6]) == size(u0)
@test size(results[7]) == size(u0)
@test size(results[8]) == size(u0)
#type tests
@test typeof(results[5]) == typeof(u0)
@test typeof(results[6]) == typeof(u0)
@test_broken typeof(results[7]) == typeof(dual)
@test_broken typeof(results[8]) == typeof(dual)
#eltype tests
@test eltype(results[5]) == eltype(u0)
@test eltype(results[7]) == eltype(dual)

#LArray tests
chunk_size = 4
u0 = LArray((2, 2); a = 1.0, b = 1.0, c = 1.0, d = 1.0)
zerodual = zero(ForwardDiff.Dual{ForwardDiff.Tag{nothing, Float64}, Float64,
                                 chunk_size})
dual = LArray((2, 2); a = zerodual, b = zerodual, c = zerodual, d = zerodual)
results = test(u0, dual, chunk_size)
#allocation tests
@test results[1] == 0
@test results[2] == 0
@test_broken results[3] == 0
@test_broken results[4] == 0
#size tests
@test size(results[5]) == size(u0)
@test size(results[6]) == size(u0)
@test size(results[7]) == size(u0)
@test size(results[8]) == size(u0)
#type tests
@test typeof(results[5]) == typeof(u0)
@test typeof(results[6]) == typeof(u0)
@test typeof(results[7]) == typeof(dual)
@test typeof(results[8]) == typeof(dual)
#eltype tests
@test eltype(results[5]) == eltype(u0)
@test eltype(results[7]) == eltype(dual)

#ArrayPartition tests
chunk_size = 2
u0 = ArrayPartition(ones(2, 2), ones(3, 3))
dual_a = zeros(ForwardDiff.Dual{ForwardDiff.Tag{nothing, Float64}, Float64,
                                chunk_size}, 2, 2)
dual_b = zeros(ForwardDiff.Dual{ForwardDiff.Tag{nothing, Float64}, Float64,
                                chunk_size}, 3, 3)
dual = ArrayPartition(dual_a, dual_b)
results = test(u0, dual, chunk_size)
#allocation tests
@test results[1] == 0
@test results[2] == 0
@test_broken results[3] == 0
@test_broken results[4] == 0
#size tests
@test size(results[5]) == size(u0)
@test size(results[6]) == size(u0)
@test size(results[7]) == size(u0)
@test size(results[8]) == size(u0)
#type tests
@test typeof(results[5]) == typeof(u0)
@test typeof(results[6]) == typeof(u0)
@test typeof(results[7]) == typeof(dual)
@test typeof(results[8]) == typeof(dual)
#eltype tests
@test eltype(results[5]) == eltype(u0)
@test eltype(results[7]) == eltype(dual)
